import os
from unittest import TestCase

from pykotor.common.module import Module
from pykotor.resource.type import ResourceType

from pykotor.common.misc import filepath_info
from pykotor.extract.installation import Installation


class TestModule(TestCase):
    def test_module(self):
        installation = Installation("C:/Program Files (x86)/Steam/steamapps/common/swkotor/")
        module = Module("danm13", installation)
        module.layout.resource()
