from unittest import TestCase

from pykotor.common.language import LocalizedString
from pykotor.common.misc import EquipmentSlot, Game
from pykotor.resource.formats.gff import load_gff
from pykotor.resource.generics.are import construct_are, dismantle_are

TEST_FILE = "../../files/test.are"


class TestARE(TestCase):
    def test_change_texture(self):
        ...

